"""TcEx Framework Module"""
# standard library
import os
from abc import ABC
from collections.abc import Generator
from textwrap import TextWrapper

# third-party
from pydantic import ValidationError
from requests import Session
from requests.exceptions import ProxyError

# first-party
from tcex.api.tc.v3._gen.model import PropertyModel
from tcex.input.field_type.sensitive import Sensitive
from tcex.pleb.cached_property import cached_property
from tcex.requests_tc.auth.hmac_auth import HmacAuth
from tcex.util import Util
from tcex.util.render.render import Render
from tcex.util.string_operation import SnakeString


class GenerateABC(ABC):
    """Generate Abstract Base Class"""

    def __init__(self, type_: SnakeString):
        """Initialize instance properties."""
        self.type_ = type_

        # properties
        self._api_server = os.getenv('TC_API_PATH')
        self.api_url: str
        self.i1 = ' ' * 4  # indent level 1
        self.i2 = ' ' * 8  # indent level 2
        self.i3 = ' ' * 12  # indent level 3
        self.i4 = ' ' * 16  # indent level 4
        self.i5 = ' ' * 20  # indent level 5
        self.messages = []
        self.requirements = {}
        self.util = Util()

    @staticmethod
    def _format_description(arg: str, description: str, length: int, indent: str) -> str:
        """Format description for field."""
        # fix descriptions coming from core API endpoint
        if description[-1] not in ('.', '?', '!'):
            description += '.'

        line = f'{arg}: {description}'
        textwrapper = TextWrapper(
            subsequent_indent=indent + '    ',
            width=length - len(indent),
            expand_tabs=True,
            tabsize=len(indent),
            break_long_words=False,
        )
        return '\n'.join(textwrapper.wrap(line))

    def _type_map(self, type_):
        """Return modified type."""
        # special handling of attributes
        if type_ == 'attributes' and self.type_ == 'cases':
            type_ = 'case_attributes'
        elif type_ == 'attributes' and self.type_ == 'groups':
            type_ = 'group_attributes'
        elif type_ == 'attributes' and self.type_ == 'indicators':
            type_ = 'indicator_attributes'
        elif type_ == 'attributes' and self.type_ == 'victims':
            type_ = 'victim_attributes'
        return self.util.snake_string(type_)

    def _module_import_data(self, type_: SnakeString) -> dict[str, str]:
        """Return the model module map data.

        This method provides the logic to build the import module and class dynamically. Using
        the provided type value (e.g., cases, groups, victim_attributes) the module
        (tcex.api.tc.v3.groups.group_filter) and the class (GroupFilter) are returned.

        filter -> the filter Class (CaseFilter)
        model  -> the model and model collection Classes (CaseModel, CasesModel)
        object -> the object and object collection Classes (Case, Cases)
        """
        type_ = self._type_map(type_)

        _base_path = f'{self.tap(type_)}.{type_}'
        return {
            'filter_module': f'{_base_path}.{type_.singular()}_filter',
            'filter_class': f'{type_.pascal_case()}Filter',
            'model_module': f'{_base_path}.{type_.singular()}_model',
            'model_class': f'{type_.singular().pascal_case()}Model',
            'model_collection_class': f'{type_.plural().pascal_case()}Model',
            'object_module': f'{_base_path}.{type_.singular()}',
            'object_class': f'{type_.singular().pascal_case()}',
            'object_collection_class': f'{type_.plural().pascal_case()}',
        }

    @cached_property
    def _prop_contents(self) -> dict:
        """Return defined API endpoint properties for the current type."""
        _properties = {}
        try:
            r = self.session.options(self.api_url, params={'show': 'readOnly'})
            # print(r.request.method, r.request.url, r.text)
            if r.ok:
                _properties = r.json()
        except (ConnectionError, ProxyError) as ex:
            Render.panel.failure(f'Failed getting types properties ({ex}).')

        return _properties

    def _prop_contents_data(self, properties: dict) -> Generator:
        """Yield the appropriate data object.

        artifacts": {
            "data": [
                {
                    "description": "a list of Artifacts corresponding to the Case",
                    "type": "Artifact"
                }
            ]
        },
        "assignee": {
            "description": "the user or group Assignee object for the Case",
            "type": "Assignee"
        },
        "attributeTypeMappings": [
            {
                "description": "the mapping of attribute types allowed for indicator types",
                "type": "AttributeTypeMappingApiInt"
            }
        ],
        """
        for field_name, field_data in sorted(properties.items()):
            if isinstance(field_data, dict):
                if 'data' in field_data and isinstance(field_data['data'], list):
                    # some properties have a data key with an array of items.
                    field_data = field_data['data'][0]

                    # if there is a data array, then the type should always be plural.
                    field_data['type'] = self.util.camel_string(field_data['type']).plural()
            elif isinstance(field_data, list):
                # in a few instance like attributeType the value of the properties key/value
                # pair is a list (currently the list only contains a single dict). to be safe
                # we loop over the list and update the type for each item.
                field_data = field_data[0]
            else:
                raise RuntimeError(
                    f'Invalid type properties data: field-name={field_name}, type={self.type_}'
                )

            field_data['name'] = field_name
            yield field_data

    @property
    def _prop_contents_updated(self) -> dict:
        """Update the properties contents, fixing issues in core data."""
        _properties = self._prop_contents

        # add id field, if missing
        self._prop_content_add_id(_properties)

        # add security label field, if missing
        self._prop_content_add_security_labels(_properties)

        # add webLink field, if missing
        self._prop_content_add_web_link(_properties)

        # fix types
        self._prop_content_fix_types(_properties)

        # remove unused fields, if any
        self._prop_content_remove_unused(_properties)

        # update "bad" data
        self._prop_content_update(_properties)

        # maybe a temp issue?
        if self.type_ == 'indicators':
            _properties['enrichment']['data'][0]['type'] = 'Enrichment'

        # critical fix for breaking API change
        if self.type_ in [
            'case_attributes',
            'group_attributes',
            'indicator_attributes',
            'victim_attributes',
        ]:
            title = 'Attribute Source Type'
            if _properties['source']['type'] != 'String':
                _properties['source']['type'] = 'String'
                self.messages.append(f'- [{self.type_}] - ({title}) - fix required.')
            else:
                self.messages.append(f'- [{self.type_}] - {title} - fix NOT required.')

        if self.type_ == 'groups':
            title = 'Filename Max Length'
            if _properties['fileName']['maxLength'] != 255:
                _properties['fileName']['maxLength'] = 255
                self.messages.append(f'- [{self.type_}] - ({title}) - fix required.')
            else:
                self.messages.append(f'- [{self.type_}] - ({title}) - fix NOT required.')

            title = 'Down Vote Count'
            if 'downVoteCount' in _properties and _properties['downVoteCount']['type'] == 'String':
                _properties['downVoteCount']['type'] = 'Integer'
                self.messages.append(f'- [{self.type_}] - ({title}) - fix required.')
            else:
                self.messages.append(f'- [{self.type_}] - ({title}) - fix NOT required.')

            title = 'Up Vote Count'
            if 'upVoteCount' in _properties and _properties['upVoteCount']['type'] == 'String':
                _properties['upVoteCount']['type'] = 'Integer'
                self.messages.append(f'- [{self.type_}] - ({title}) - fix required.')
            else:
                self.messages.append(f'- [{self.type_}] - ({title}) - fix NOT required.')

        if self.type_ in [
            'cases',
            'tasks',
            'workflow_templates',
        ]:
            if _properties.get('owner'):
                title = 'Read Only Owner Fix'
                if _properties['owner'].get('readOnly') is not True:
                    _properties['owner']['readOnly'] = True
                    self.messages.append(f'- [{self.type_}] - ({title}) - fix required.')
                else:
                    self.messages.append(f'- [{self.type_}] - ({title}) - fix NOT required.')

                title = 'Updatable Owner Fix'
                if _properties['owner'].get('updatable') is not False:
                    _properties['owner']['updatable'] = False
                    self.messages.append(f'- [{self.type_}] - ({title}) - fix required.')
                else:
                    self.messages.append(f'- [{self.type_}] - ({title}) - fix NOT required.')

            if _properties.get('ownerId'):
                title = 'Read Only Owner Id Fix'
                if _properties['ownerId'].get('readOnly') is not True:
                    _properties['ownerId']['readOnly'] = True
                    self.messages.append(f'- [{self.type_}] - ({title}) - fix required.')
                else:
                    self.messages.append(f'- [{self.type_}] - ({title}) - fix NOT required.')

                title = 'Updatable Owner Id Fix'
                if _properties['ownerId'].get('updatable') is not False:
                    _properties['ownerId']['updatable'] = False
                    self.messages.append(f'- [{self.type_}] - ({title}) - fix required.')
                else:
                    self.messages.append(f'- [{self.type_}] - ({title}) - fix NOT required.')

        return _properties

    def _prop_content_add_id(self, properties: dict):
        """Add id field to properties.

        For some reason the core API does not return the id field for some types.
        """
        if 'id' not in properties:
            properties['id'] = {
                'required': False,
                'type': 'Integer',
                'description': 'The id of the **Object**',
                'readOnly': True,
                'updatable': False,
            }
            self.messages.append(f'- [{self.type_}] - The id field is missing.')

    def _prop_content_add_security_labels(self, properties: dict):
        """Add securityLabels field to properties."""
        if self.type_ in ['group_attributes', 'indicator_attributes']:
            if 'securityLabels' not in properties:
                properties['securityLabels'] = {
                    'data': [
                        {
                            'description': (
                                'A list of Security Labels corresponding to the Intel item '
                                '(NOTE: Setting this parameter will replace any existing '
                                'tag(s) with the one(s) specified)'
                            ),
                            'maxSize': 1000,
                            'required': False,
                            'type': 'SecurityLabel',
                        }
                    ]
                }
                self.messages.append(f'- [{self.type_}] - The securityLabels field is missing.')
            else:
                self.messages.append(f'- [{self.type_}] - The securityLabels field is NOT missing.')

    def _prop_content_add_web_link(self, properties: dict):
        """Add webLink field to properties.

        For some reason the core API does not return the webLink field for some types.
        """
        if self.type_ in ['groups', 'indicators']:
            if 'webLink' not in properties:
                properties['webLink'] = {
                    'required': False,
                    'type': 'String',
                    'description': 'The object link.',
                    'readOnly': True,
                    'updatable': False,
                }
                self.messages.append(f'- [{self.type_}] - The webLink field is missing.')
            else:
                self.messages.append(f'- [{self.type_}] - The webLink field is NOT missing.')

    def _prop_content_fix_types(self, _properties: dict):
        """Return modified type."""
        type_ = _properties.get('type')

        if type_ == 'attributes' and self.type_ == 'cases':
            _properties['type'] = 'caseAttributes'
        elif type_ == 'attributes' and self.type_ == 'groups':
            _properties['type'] = 'groupAttributes'
        elif type_ == 'attributes' and self.type_ == 'indicators':
            _properties['type'] = 'indicatorAttributes'
        elif type_ == 'attributes' and self.type_ == 'victims':
            _properties['type'] = 'victimAttributes'

    def _prop_content_remove_unused(self, properties: dict):
        """Remove unused fields from properties."""
        if self.type_ in ['attribute_types']:
            if 'owner' in properties:
                del properties['owner']

        if self.type_ in [
            'attribute_types',
            'case_attributes',
            'group_attributes',
            'indicator_attributes',
            'victim_attributes',
        ]:
            unused_fields = [
                'attributeTypeMappings',
                'ownerId',  # Core Issue: should not be listed for this type
                'ownerName',  # Core Issue: should not be listed for this type
                'settings',
                'tags',  # Core Issue: should not be listed for this type
                'webLink',  # Core Issue: should not be listed for this type
            ]
            for field in unused_fields:
                if field in properties:
                    del properties[field]

        if self.type_ in ['users']:
            unused_fields = [
                'customTqlTimeout',
                'disabled',
                'firstName',
                'jobFunction',
                'jobRole',
                'lastLogin',
                'lastName',
                'lastPasswordChange',
                'locked',
                'logoutIntervalMinutes',
                'owner',
                'ownerRoles',
                'password',
                'passwordResetRequired',
                'pseudonym',
                'systemRole',
                'termsAccepted',
                'termsAcceptedDate',
                'tqlTimeout',
                'twoFactorResetRequired',
                'uiTheme',
            ]
            for field in unused_fields:
                if field in properties:
                    del properties[field]

        if self.type_ in ['victims']:
            unused_fields = [
                'ownerId',  # Core Issue: should not be listed for this type
            ]
            for field in unused_fields:
                if field in properties:
                    del properties[field]

    def _prop_content_update(self, properties: dict):
        """Update "bad" data in properties."""
        if self.type_ in ['groups']:
            # fixed fields that are missing readOnly property
            if 'downVoteCount' in properties:
                properties['downVoteCount']['readOnly'] = True
            if 'upVoteCount' in properties:
                properties['upVoteCount']['readOnly'] = True

        if self.type_ in ['victims']:
            # ownerName is readOnly, but readOnly is not defined in response from OPTIONS endpoint
            properties['ownerName']['readOnly'] = True

    @property
    def _prop_models(self) -> list[PropertyModel]:
        """Return a list of PropertyModel objects."""
        properties_models = []
        for field_data in self._prop_contents_data(self._prop_contents_updated):
            try:
                properties_models.append(PropertyModel(**field_data))
            except ValidationError as ex:
                # print(field_data)
                Render.panel.failure(
                    f'Failed generating property model: data={field_data} ({ex}).',
                )
        return properties_models

    def gen_requirements(self):
        """Generate imports string."""
        # add additional imports when required
        if self.requirements.get('type-checking'):
            self.requirements['standard library'].append('from typing import TYPE_CHECKING')

        indent = ''
        # _libs: list[dict | str] = []
        _libs = []
        for from_, libs in self.requirements.items():
            if not libs:
                # continue if there are no libraries to import
                continue

            if from_ in ['first-party-forward-reference']:
                # skip forward references
                continue

            comment = ''
            if from_ == 'type-checking':
                _libs.append('if TYPE_CHECKING:  # pragma: no cover')
                indent = self.i1
                # this should be fine?
                from_ = 'first-party'
                comment = '  # CIRCULAR-IMPORT'

            _libs.append(f'{indent}# {from_}')

            # manage imports as string and dicts
            _imports = []  # temp store for imports so they can be sorted
            for lib in libs:
                if isinstance(lib, dict):
                    imports = ', '.join(sorted(lib.get('imports')))  # type: ignore
                    _imports.append(
                        f'''{indent}from {lib.get('module')} import {imports}{comment}'''
                    )
                elif isinstance(lib, str):
                    _imports.append(f'{indent}{lib}{comment}')
            _libs.extend(sorted(_imports))  # add imports sorted

            _libs.append('')  # add newline
        _libs.append('')  # add newline

        # This is the last part of the requirements generated.
        return '\n'.join(_libs)  # type: ignore

    @property
    def session(self) -> Session:
        """Return Session configured for TC API."""
        _session = Session()
        _session.auth = HmacAuth(
            os.getenv('TC_API_ACCESS_ID', ''), Sensitive(os.getenv('TC_API_SECRET_KEY', ''))
        )
        return _session

    def tap(self, type_: str):
        """Return the TcEx Api Path."""
        type_ = self.util.snake_string(type_)
        if type_.plural().lower() in [
            'owners',
            'owner_roles',
            'system_roles',
            'users',
            'user_groups',
        ]:
            return 'tcex.api.tc.v3.security'
        return 'tcex.api.tc.v3'



"""TcEx Framework Module"""
# pylint: disable=no-member,no-self-argument,wrong-import-position
# standard library
from datetime import datetime

# third-party
from pydantic import BaseModel, Extra, Field, PrivateAttr

# first-party
from tcex.api.tc.v3.v3_model_abc import V3ModelABC
from tcex.util import Util


class SecurityLabelModel(
    V3ModelABC,
    alias_generator=Util().snake_to_camel,
    extra=Extra.allow,
    title='SecurityLabel Model',
    validate_assignment=True,
):
    """Security_Label Model"""

    _associated_type = PrivateAttr(False)
    _cm_type = PrivateAttr(False)
    _shared_type = PrivateAttr(True)
    _staged = PrivateAttr(False)

    color: str | None = Field(
        None,
        description='Color of the security label.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='color',
    )
    date_added: datetime | None = Field(
        None,
        allow_mutation=False,
        description='The date and time that the label was added.',
        read_only=True,
        title='dateAdded',
    )
    description: str | None = Field(
        None,
        description='Description of the security label.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='description',
    )
    id: int | None = Field(
        None,
        description='The ID of the item.',
        read_only=True,
        title='id',
    )
    name: str | None = Field(
        None,
        description='Name of the security label.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='name',
    )
    owner: str | None = Field(
        None,
        description='The name of the Owner of the Label.',
        methods=['POST'],
        read_only=False,
        title='owner',
    )


class SecurityLabelDataModel(
    BaseModel,
    title='SecurityLabel Data Model',
    alias_generator=Util().snake_to_camel,
    validate_assignment=True,
):
    """Security_Labels Data Model"""

    data: list[SecurityLabelModel] | None = Field(
        [],
        description='The data for the SecurityLabels.',
        methods=['POST', 'PUT'],
        title='data',
    )


class SecurityLabelsModel(
    BaseModel,
    title='SecurityLabels Model',
    alias_generator=Util().snake_to_camel,
    validate_assignment=True,
):
    """Security_Labels Model"""

    _mode_support = PrivateAttr(True)

    data: list[SecurityLabelModel] | None = Field(
        [],
        description='The data for the SecurityLabels.',
        methods=['POST', 'PUT'],
        title='data',
    )
    mode: str = Field(
        'append',
        description='The PUT mode for nested objects (append, delete, replace). Default: append',
        methods=['POST', 'PUT'],
        title='append',
    )


# add forward references
SecurityLabelDataModel.update_forward_refs()
SecurityLabelModel.update_forward_refs()
SecurityLabelsModel.update_forward_refs()



"""TcEx Framework Module"""
# standard library
from collections.abc import Iterator

# first-party
from tcex.api.tc.v3.api_endpoints import ApiEndpoints
from tcex.api.tc.v3.object_abc import ObjectABC
from tcex.api.tc.v3.object_collection_abc import ObjectCollectionABC
from tcex.api.tc.v3.security.system_roles.system_role_filter import SystemRoleFilter
from tcex.api.tc.v3.security.system_roles.system_role_model import SystemRoleModel, SystemRolesModel


class SystemRole(ObjectABC):
    """SystemRoles Object."""

    def __init__(self, **kwargs):
        """Initialize instance properties."""
        super().__init__(kwargs.pop('session', None))

        # properties
        self._model: SystemRoleModel = SystemRoleModel(**kwargs)
        self._nested_field_name = 'systemRoles'
        self._nested_filter = 'has_system_role'
        self.type_ = 'System Role'

    @property
    def _api_endpoint(self) -> str:
        """Return the type specific API endpoint."""
        return ApiEndpoints.SYSTEM_ROLES.value

    @property
    def model(self) -> SystemRoleModel:
        """Return the model data."""
        return self._model

    @model.setter
    def model(self, data: dict | SystemRoleModel):
        """Create model using the provided data."""
        if isinstance(data, type(self.model)):
            # provided data is already a model, nothing required to change
            self._model = data
        elif isinstance(data, dict):
            # provided data is raw response, load the model
            self._model = type(self.model)(**data)
        else:
            raise RuntimeError(f'Invalid data type: {type(data)} provided.')


class SystemRoles(ObjectCollectionABC):
    """SystemRoles Collection.

    # Example of params input
    {
        'result_limit': 100,  # Limit the retrieved results.
        'result_start': 10,  # Starting count used for pagination.
        'fields': ['caseId', 'summary']  # Select additional return fields.
    }

    Args:
        session (Session): Session object configured with TC API Auth.
        tql_filters (list): List of TQL filters.
        params (dict): Additional query params (see example above).
    """

    def __init__(self, **kwargs):
        """Initialize instance properties."""
        super().__init__(
            kwargs.pop('session', None), kwargs.pop('tql_filter', None), kwargs.pop('params', None)
        )
        self._model = SystemRolesModel(**kwargs)
        self.type_ = 'system_roles'

    def __iter__(self) -> Iterator[SystemRole]:
        """Return CM objects."""
        return self.iterate(base_class=SystemRole)  # type: ignore

    @property
    def _api_endpoint(self) -> str:
        """Return the type specific API endpoint."""
        return ApiEndpoints.SYSTEM_ROLES.value

    @property
    def filter(self) -> SystemRoleFilter:
        """Return the type specific filter object."""
        return SystemRoleFilter(self.tql)



"""TcEx Framework Module"""
# standard library
from datetime import datetime
from enum import Enum

# third-party
from arrow import Arrow

# first-party
from tcex.api.tc.v3.api_endpoints import ApiEndpoints
from tcex.api.tc.v3.filter_abc import FilterABC
from tcex.api.tc.v3.tql.tql import Tql
from tcex.api.tc.v3.tql.tql_operator import TqlOperator
from tcex.api.tc.v3.tql.tql_type import TqlType


class NoteFilter(FilterABC):
    """Filter Object for Notes"""

    @property
    def _api_endpoint(self) -> str:
        """Return the API endpoint."""
        return ApiEndpoints.NOTES.value

    def artifact_id(self, operator: Enum, artifact_id: int | list):
        """Filter Artifact ID based on **artifactId** keyword.

        Args:
            operator: The operator enum for the filter.
            artifact_id: The ID of the artifact this note is associated with.
        """
        if isinstance(artifact_id, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('artifactId', operator, artifact_id, TqlType.INTEGER)

    def author(self, operator: Enum, author: list | str):
        """Filter Author based on **author** keyword.

        Args:
            operator: The operator enum for the filter.
            author: The account login of the user who wrote the note.
        """
        if isinstance(author, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('author', operator, author, TqlType.STRING)

    def case_id(self, operator: Enum, case_id: int | list):
        """Filter Case ID based on **caseId** keyword.

        Args:
            operator: The operator enum for the filter.
            case_id: The ID of the case this note is associated with.
        """
        if isinstance(case_id, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('caseId', operator, case_id, TqlType.INTEGER)

    def date_added(self, operator: Enum, date_added: Arrow | datetime | int | str):
        """Filter Date Added based on **dateAdded** keyword.

        Args:
            operator: The operator enum for the filter.
            date_added: The date the note was written.
        """
        date_added = self.util.any_to_datetime(date_added).strftime('%Y-%m-%d %H:%M:%S')
        self._tql.add_filter('dateAdded', operator, date_added, TqlType.STRING)

    @property
    def has_artifact(self):
        """Return **ArtifactFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.artifacts.artifact_filter import ArtifactFilter

        artifacts = ArtifactFilter(Tql())
        self._tql.add_filter('hasArtifact', TqlOperator.EQ, artifacts, TqlType.SUB_QUERY)
        return artifacts

    @property
    def has_case(self):
        """Return **CaseFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.cases.case_filter import CaseFilter

        cases = CaseFilter(Tql())
        self._tql.add_filter('hasCase', TqlOperator.EQ, cases, TqlType.SUB_QUERY)
        return cases

    @property
    def has_task(self):
        """Return **TaskFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.tasks.task_filter import TaskFilter

        tasks = TaskFilter(Tql())
        self._tql.add_filter('hasTask', TqlOperator.EQ, tasks, TqlType.SUB_QUERY)
        return tasks

    def id(self, operator: Enum, id: int | list):  # pylint: disable=redefined-builtin
        """Filter ID based on **id** keyword.

        Args:
            operator: The operator enum for the filter.
            id: The ID of the case.
        """
        if isinstance(id, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('id', operator, id, TqlType.INTEGER)

    def last_modified(self, operator: Enum, last_modified: Arrow | datetime | int | str):
        """Filter Last Modified based on **lastModified** keyword.

        Args:
            operator: The operator enum for the filter.
            last_modified: The date the note was last modified.
        """
        last_modified = self.util.any_to_datetime(last_modified).strftime('%Y-%m-%d %H:%M:%S')
        self._tql.add_filter('lastModified', operator, last_modified, TqlType.STRING)

    def summary(self, operator: Enum, summary: list | str):
        """Filter Summary based on **summary** keyword.

        Args:
            operator: The operator enum for the filter.
            summary: Text of the first 100 characters of the note.
        """
        if isinstance(summary, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('summary', operator, summary, TqlType.STRING)

    def task_id(self, operator: Enum, task_id: int | list):
        """Filter Task ID based on **taskId** keyword.

        Args:
            operator: The operator enum for the filter.
            task_id: The ID of the task this note is associated with.
        """
        if isinstance(task_id, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('taskId', operator, task_id, TqlType.INTEGER)

    def workflow_event_id(self, operator: Enum, workflow_event_id: int | list):
        """Filter Workflow Event ID based on **workflowEventId** keyword.

        Args:
            operator: The operator enum for the filter.
            workflow_event_id: The ID of the workflow event this note is associated with.
        """
        if isinstance(workflow_event_id, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('workflowEventId', operator, workflow_event_id, TqlType.INTEGER)



"""TcEx Framework Module"""
# standard library
from typing import TYPE_CHECKING

# first-party
from tcex.api.tc.v2.threat_intelligence.mapping.indicator.indicator import Indicator

if TYPE_CHECKING:
    # first-party
    from tcex.api.tc.v2.threat_intelligence.threat_intelligence import ThreatIntelligence


class EmailAddress(Indicator):
    """Unique API calls for Email Address API Endpoints"""

    def __init__(self, ti: 'ThreatIntelligence', **kwargs):
        """Initialize instance properties.

        Args:
            ti (ThreatIntelligence): An instance of the ThreatIntelligence Class.
            **kwargs: Additional keyword arguments.

        Keyword Args:
            address (str): The value for this Indicator.
            active (bool, kwargs): If False the indicator is marked "inactive" in TC.
            confidence (str, kwargs): The threat confidence for this Indicator.
            date_added (str, kwargs): [Read-Only] The date timestamp the Indicator was created.
            last_modified (str, kwargs): The date timestamp the Indicator was last modified.
            private_flag (bool, kwargs): If True the indicator is marked as private in TC.
            rating (str, kwargs): The threat rating for this Indicator.
            xid (str, kwargs): The external id for this Indicator.
        """
        super().__init__(
            ti,
            sub_type='Email Address',
            api_entity='emailAddress',
            api_branch='emailAddresses',
            **kwargs,
        )
        self.unique_id = kwargs.get('unique_id', kwargs.get('address'))
        self._data['address'] = self.unique_id

    def can_create(self):
        """Return True if email address can be created.

        If the address has been provided returns that the EmailAddress can be created, otherwise
        returns that the EmailAddress cannot be created.
        """
        return self.data.get('address') is not None

    def _set_unique_id(self, json_response: dict):
        """Set the unique_id provided a json response."""
        self.unique_id = json_response.get('address', '')



"""TcEx Framework Module"""
# pylint: disable=no-member,no-self-argument,wrong-import-position

# third-party
from pydantic import Field

# first-party
from tcex.api.tc.v3.security.users.user_model import UserModel
from tcex.util import Util


class AssigneeUserModel(
    UserModel,
    title='Assignee User Model',
    alias_generator=Util().snake_to_camel,
    validate_assignment=True,
):
    """Assignee Model"""

    user_name: str | None = Field(
        None,
        allow_mutation=False,
        description='The **user name** for the User.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='userName',
    )



"""TcEx Framework Module"""
# standard library
from collections.abc import Callable
from typing import Any, cast

# third-party
import wrapt

# first-party
from tcex.exit import ExitCode
from tcex.tcex import TcEx


class FailOnOutput:
    """Fail App if return value (output) value conditions are met.

    This decorator allows for the App to exit on conditions defined in the function
    parameters.

    .. code-block:: python
        :linenos:
        :lineno-start: 1

        @FailOnOutput(
            fail_on=['false'], fail_msg='Operation returned a value of "false".'
        )
        def my_method(data):
            return data.lowercase()

    Args:
        fail_enabled: Accepts a boolean or string value.  If a boolean value
            is provided that value will control enabling/disabling this feature. A string
            value should reference an item in the args namespace which resolves to a boolean.
            The value of this boolean will control enabling/disabling this feature.
        fail_msg: The message to log when raising RuntimeError.
        fail_msg_property: The App property containing the dynamic exit message.
        fail_on: Fail if return value from App method is in the list.
        write_output: Defaults to True. If true, will call App.write_outputs() before failing
            on matched fail_on value.
    """

    def __init__(
        self,
        fail_enabled: bool | str = True,
        fail_msg: str = 'Method returned invalid output.',
        fail_on: list | None = None,
        fail_msg_property: str | None = None,
        write_output: bool = True,
    ):
        """Initialize instance properties."""
        self.fail_enabled = fail_enabled
        self.fail_msg = fail_msg
        self.fail_on = fail_on or []
        self.fail_msg_property = fail_msg_property
        self.write_output = write_output

    @wrapt.decorator
    def __call__(self, *wrapped_args) -> Any:
        """Implement __call__ function for decorator.

        Args:
            wrapped: The wrapped function which in turns
                needs to be called by your wrapper function.
            instance: The object to which the wrapped
                function was bound when it was called.
            args: The list of positional arguments supplied
                when the decorated function was called.
            kwargs: The dictionary of keyword arguments
                supplied when the decorated function was called.

        Returns:
            function: The custom decorator function.
        """
        # using wrapped args to support typing hints in PyRight
        wrapped: Callable = wrapped_args[0]
        app: Any = wrapped_args[1]
        args: list = wrapped_args[2] if len(wrapped_args) > 1 else []
        kwargs: dict = wrapped_args[3] if len(wrapped_args) > 2 else {}

        def fail() -> Any:
            """Call the function and store or append return value."""
            tcex = cast(TcEx, app.tcex)

            # call method to get output
            data = wrapped(*args, **kwargs)

            # self.enable (e.g., True or 'fail_on_false') enables/disables this feature
            enabled = self.fail_enabled
            if isinstance(enabled, str):
                # get enabled value from App inputs
                enabled = getattr(tcex.inputs.model, enabled)
                if not isinstance(enabled, bool):  # pragma: no cover
                    raise RuntimeError(
                        'The fail_enabled value must be a boolean or resolved to bool.'
                    )
                tcex.log.debug(f'Fail enabled is {enabled} ({self.fail_enabled}).')

            failed = False
            if enabled:
                if isinstance(data, list):
                    # validate each value in the list of results.
                    for d in data:
                        if d in self.fail_on:
                            failed = True
                            break
                else:
                    if data in self.fail_on:
                        failed = True

                if failed:
                    if self.write_output:
                        tcex.app.playbook.output.process()
                        if hasattr(app, 'write_output'):
                            app.write_output()
                    app.exit_message = self.get_fail_msg(app)  # for test cases
                    tcex.exit.exit(ExitCode.FAILURE, self.get_fail_msg(app))
            return data

        return fail()

    def get_fail_msg(self, app) -> str:
        """Return the appropriate fail message."""
        fail_msg = self.fail_msg
        if isinstance(self.fail_msg_property, str):
            if self.fail_msg_property and hasattr(app, self.fail_msg_property):
                fail_msg = getattr(app, self.fail_msg_property)
        return fail_msg



"""TcEx Framework Module"""
# standard library
from typing import TYPE_CHECKING

# first-party
from tcex.api.tc.v2.threat_intelligence.mapping.group.group import Group

if TYPE_CHECKING:
    # first-party
    from tcex.api.tc.v2.threat_intelligence.threat_intelligence import (
        ThreatIntelligence,  # CIRCULAR-IMPORT
    )


class Report(Group):
    """Unique API calls for Report API Endpoints

    Args:
        ti: An instance of the ThreatIntelligence Class.
        name (str, kwargs): The name for this Group.
        file_name (str, kwargs): The name for the attached file for this Group.
        publish_date (str, kwargs): The publish datetime expression for this Group.
    """

    def __init__(self, ti: 'ThreatIntelligence', **kwargs):
        """Initialize instance properties."""

        super().__init__(ti, sub_type='Report', api_entity='report', api_branch='reports', **kwargs)

    def file_content(self, file_content: str, update_if_exists: bool = True):
        """Update the file content."""
        if not self.can_update():
            self._handle_error(910, [self.type])

        self._data['fileContent'] = file_content
        return self.tc_requests.upload(
            self.api_type,
            self.api_branch,
            self.unique_id,
            file_content,
            update_if_exists=update_if_exists,
        )

    def file_name(self, file_name: str):
        """Update the file_name."""
        if not self.can_update():
            self._handle_error(910, [self.type])

        self._data['fileName'] = file_name
        request = {'fileName': file_name}
        return self.tc_requests.update(self.api_type, self.api_branch, self.unique_id, request)

    def file_size(self, file_size: int):
        """Update the file_size."""
        if not self.can_update():
            self._handle_error(910, [self.type])

        self._data['fileSize'] = file_size
        request = {'fileSize': file_size, 'fileName': self._data['fileName']}
        return self.tc_requests.update(self.api_type, self.api_branch, self.unique_id, request)

    def get_file_hash(self, hash_type: str = 'sha256'):
        """Get the hash value of attached document"""
        if not self.can_update():
            self._handle_error(910, [self.type])

        return self.tc_requests.get_file_hash(
            self.api_type, self.api_branch, self.unique_id, hash_type=hash_type
        )

    def status(self, status: str):
        """Update the status

        Valid status:
        + Success
        + Awaiting Upload
        + In Progress
        + Failed

        Args:
            status: Success, Awaiting Upload, In Progress, or Failed
        """
        if not self.can_update():
            self._handle_error(910, [self.type])

        self._data['status'] = status
        request = {'status': status, 'fileName': self._data['fileName']}
        return self.tc_requests.update(self.api_type, self.api_branch, self.unique_id, request)

    def publish_date(self, publish_date):
        """Return Email to."""
        if not self.can_update():
            self._handle_error(910, [self.type])

        publish_date = self.util.any_to_datetime(publish_date).strftime('%Y-%m-%dT%H:%M:%SZ')

        self._data['publishDate'] = publish_date
        request = {'publishDate': publish_date, 'fileName': self._data['fileName']}
        return self.tc_requests.update(self.api_type, self.api_branch, self.unique_id, request)

    def download(self):
        """Download the documents context."""
        if not self.can_update():
            self._handle_error(910, [self.type])

        return self.tc_requests.download(self.api_type, self.api_branch, self.unique_id)
